; Copyright (c) 2006, Theo Brauers, Forschungszentrum Juelich GmbH ICG-II
; All rights reserved. Unauthorized reproduction prohibited.
; This software may be used, copied, or redistributed as long as it is not
; sold and this copyright notice is reproduced on each copy made.  This
; routine is provided as is without any express or implied warranties
; whatsoever. Email bug reports to th.brauers@fz-juelich.de
;
;+
; USERLEVEL:
;   TOPLEVEL
;
; NAME:
;  	check_edf
;
; PURPOSE:
;   This function reads an eurochamp data file (edf) and checks the compatibility.
;
; CATEGORY:
;   DATAFILES/FILE
;
; CALLING SEQUENCE:
;   result = check_edf(edf_file)
;
; INPUTS:
;   edf_file: Name of the edf file. If omitted an open file dialog box appears.
;
; KEYWORD PARAMETERS:
;   ERROR = err: If set to a named variable the error status (0: no error) of the routine is returned.
;
; OUTPUTS:
;   This function returns a text arry as areport of the checks.
;
; RESTRICTIONS:
;   This function checks if the format of the edf file obeys the rules published
;	at http://saphir.fz-juelich.de/eurochamp/
;
; EXAMPLE:
;   result = check_edf(edf_file, ERROR=err)
;
; MODIFICATION HISTORY:
;   Written by: Theo Brauers 2007-1-5 (based on read_enz)
;   Modified by: Theo Brauers 2009-10-9: bug in line 113, idg --> idp
;-
FUNCTION check_edf, file, ERROR = err, _EXTRA =e

; this block defines the parameters neded in the header
; it could also be read from file if required
;
required_global = ['PI_NAME','EMAIL', 'TITLE','INSTITUTE' $
	, 'DATA_CATEGORY','TYPE_OF_DATA', 'STATUS_OF_FILE' $
	, 'PLATFORM','NAME_OF_PLATFORM']
allowed_global =['PHONE', 'WORKGROUP', 'EXPERIMENT','CAMPAIGN'$
	, 'VERSION', 'SOURCE_FILE', 'DESCRIPTION']
required_time =['LONG_NAME', 'UNITS']
allowed_time = ['SHORT_NAME']
required_param =['LONG_NAME','UNITS']
allowed_param =['DESCRIPTION','SHORT_NAME','MISSING_VALUE' $
	,'ACCURACY_ABS','ACCURACY_REL','INSTRUMENT' $
	,'CALIBRATION','INSTRUMENT_FILE','ANALYSIS_FILE']

; set error must be 1 for all exits befor the program ended
err = 1

;write startup message to output variable
result = 'FILE:= '+ file

; test file for access as reguar file
IF NOT FILE_TEST(file, /REG, /READ) THEN RETURN, [result, 'ERROR: cannot access file']

; read ascii file into string array
x = get_file(file, ERROR = get_file_err)
IF get_file_err NE 0 THEN RETURN, [result, 'ERROR:= cannot read file (get_file.pro)']
result = [result, 'STATUS:= file succesfully read (get_file.pro)']

; get rid of comments, leading and trailing blanks
nx1 = N_ELEMENTS(x)
x = drop_comments(x, IGNORE='!')
nx2 = N_ELEMENTS(x)
IF nx2 LT nx1 THEN  result = [result, 'STATUS:= '+n2s(nx1-nx2)+' lines of comment removed (drop_comments.pro)']
x = STRCOMPRESS(STRTRIM(x,2))
result = [result, 'STATUS:= leading and trailing blanks removed (STRTRIM)']
result = [result, 'STATUS:= tabs converted to single spaces (STRCOMPRESS)']

; spearator required
ida  = WHERE(STRPOS(x,'&&&&&') EQ 0, cda)
IF cda EQ 0 THEN RETURN, [result, 'ERROR:= missing  &&&&&  detected']
IF cda GT 1 THEN RETURN, [result, 'ERROR:= multiple  &&&&&  detected ('+n2s(cda)+')']
result = [result, 'STATUS:= exactly 1 line with  &&&&&  found ']

; check header and data sizes
IF ida[0] EQ 0 THEN RETURN, [result, 'ERROR:= no header detected']
IF ida[0] GE N_ELEMENTS(x)-1 THEN RETURN, [result, file + 'ERROR:= no data detected']

; get header and data in separate arrays
h = STRTRIM(STRCOMPRESS(x[0:ida[0]-1]))
d = STRTRIM(STRCOMPRESS(x[ida[0]+1:*]))
result = [result, 'HEADER_RECORDS:= '+ n2s(N_ELEMENTS(h))]
result = [result, 'DATA_RECORDS:= '+ n2s(N_ELEMENTS(d))]

; get rid of "
idx2 = WHERE(STRPOS(h,'"') EQ 0 AND  STRPOS(h,'"', /REVERSE_SEARC) EQ STRLEN(h)-1, cdx2)
IF cdx2 GT 0 THEN BEGIN
	FOR jj = 0L, cdx2-1 DO h[idx2[jj]] = STRMID(h[idx2[jj]],1,STRLEN(h[idx2[jj]])-2 )
	result = [result, 'WARNING:= removed " , please submit file w/o "']
ENDIF

; processing header
idx = WHERE(STRPOS(h,'X_HEADER=') GE 0,cdx)
IF cdx LT 3 OR cdx GT 4 THEN RETURN, [result, 'ERROR:= wrong number of X_HEADER= ... entries']

idg = WHERE(STRPOS(h,'X_HEADER=NETCDF_GLOBAL') GE 0,cdg)
IF cdg NE 1 THEN RETURN, [result, 'ERROR:= X_HEADER=NETCDF_GLOBAL missing']
idp = WHERE(STRPOS(h,'X_HEADER=NETCDF_PARAMETER') GE 0,cdp)
IF cdp NE 1 THEN RETURN, [result,  'ERROR:= X_HEADER=NETCDF_PARAMETER missing']
ide = WHERE(STRPOS(h,'X_HEADER=ENZ') GE 0,cde)
IF cde NE 1 THEN RETURN, [result, 'ERROR:= X_HEADER=ENZ missing']

; search the number of columns entry
idn = WHERE(STRPOS(h,'NUMBER OF COLUMNS=') GE 0,cdn)
IF cdn NE 1 THEN RETURN, [result, 'ERROR:= NUMBER OF COLUMNS entry missing']
ncol = LONG(str_sep2(h[idn[0]], '=', ONLY=1))
IF ncol LE 0 THEN RETURN, [result, 'ERROR:= NUMBER OF COLUMNS equal ZERO found, no data specified']
IF ncol EQ 1 THEN RETURN, [result, 'ERROR:= NUMBER OF COLUMNS equal ONE found, one data column ??']
result = [result, 'STATUS:= NUMBER_OF_COLUMNS='+n2s(ncol)]

; loop all lines with COLUMNnn= in the enz section
text ='' & num = -1L
FOR ii = ide[0] + 1L, N_ELEMENTS(h)-1 DO BEGIN
	IF STRPOS(h[ii],'X_HEADER=') GE 0 THEN BREAK
	IF STRPOS(h[ii],'COLUMN') NE 0 THEN CONTINUE
	IF STRPOS(h[ii],'=') LT 5 THEN CONTINUE
	w = str_sep2(h[ii], '=')
	text = [text, w[1]]
	num = [num, LONG(STRMID(w[0],6))]
ENDFOR
num = num[1:*] & text = STRTRIM(text[1:*],2)
idsrt = SORT(num)
num = num[idsrt] & text = STRUPCASE(text[idsrt])

idx_short = WHERE(STRPOS(h,'SHORT_NAME(') EQ 0, cdx_short)
IF cdx_short EQ 0 THEN $
	RETURN, [result, 'ERROR:= each data COLUMN needs a SHORT_NAME entry']
text_short = STRUPCASE(STRTRIM(str_sep2(h[idx_short],'=', ONLY=1),2))

num_sd = MAKE_ARRAY(ncol, VAL=-1L)
IF TOTAL(LINDGEN(ncol)+1 EQ num) NE ncol THEN $
	RETURN, [result, 'ERROR:= COLUMN entries missing']
FOR ii = 0L, ncol-1 DO $
	result = [result, 'STATUS:= COLUMN '+n2s(num[ii])+'='+text[ii]]
FOR ii = 0L, ncol-1 DO BEGIN
	IF STRPOS(text[ii],'STDEV(') EQ 0 OR text[ii] EQ 'INTERVAL' THEN CONTINUE
	idx_sn = WHERE(STRUPCASE(text[ii]) EQ STRUPCASE(text_short), cdx_sn)
	CASE cdx_sn OF
		0: result = [result, 'ERROR:= no SHORT_NAME found for COLUMN '+n2s(num[ii])+'='+text[ii]]
		1: result = [result, 'STATUS:= SHORT_NAME found for COLUMN '+n2s(num[ii])+'='+text[ii]]
		ELSE: result = [result, 'ERROR:= more than one SHORT_NAME found for COLUMN '+n2s(num[ii])+'='+text[ii]]
	ENDCASE
ENDFOR

FOR ii = 0L, ncol-1 DO BEGIN
	IF STRPOS(text[ii],'STDEV(') EQ 0 THEN CONTINUE
	IF text[ii] NE text2tagname(text[ii]) THEN $
		result = [result, 'ERROR:= NOT a valid SHORT_NAME '+text[ii]]
ENDFOR
IF (WHERE(STRPOS(result, 'ERROR:=') EQ 0))[0] NE -1 THEN RETURN, result

FOR ii = 0L, ncol-1 DO BEGIN
	IF STRPOS(text[ii],'STDEV(') EQ 0 THEN BEGIN
		num_sd[ii] = -2
		IF ii EQ 0 THEN RETURN, [result, 'ERROR:= STDEV(...) must not appear in first column']
		num_sd[ii-1] = ii
		IF text[ii-1] NE STRTRIM(within_brackets(text[ii], BRA=['STDEV(',')'])) THEN $
			RETURN, [result, 'ERROR:= COLUMN '+n2s(num[ii]) $
				+' NOT detected an STDEV column of COLUMN '+n2s(num[ii-1]) $
				+ ', please check the SPELLING of both entries']
		result = [result, 'STATUS:= detected COLUMN '+n2s(num[ii]) $
			+' as an STDEV column of COLUMN '+ n2s(num[ii-1])]
	ENDIF
ENDFOR

; check if time axis is available
IF STRPOS(STRUPCASE(text[0]),'TIME') NE 0 THEN BEGIN
	result = [result, 'STATUS:= NO COLUMN detected as a TIME column, please check!']
	_time_flag = 0
ENDIF ELSE BEGIN
	_time_flag =1
	IF STRPOS(STRUPCASE(text[1]),'INTERVAL') EQ 0 THEN BEGIN
		num_sd[0] = 1 & num_sd[1] = -2
	ENDIF
	idt = WHERE(STRPOS(h,'X_HEADER=NETCDF_TIME') GE 0,cdt)
	IF cdt NE 1 THEN RETURN, [result, 'ERROR:= X_HEADER=NETCDF_TIME missing']
	result = [result, 'STATUS:= detected COLUMN '+n2s(num[0])+' as a TIME column']
	IF num_sd[0] EQ 1 THEN result = [result, 'STATUS:= detected COLUMN '+n2s(num[1])+' as an INTERVAL column']
	time = CREATE_STRUCT('LONG_NAME', '_NOTDEF_REQUIRED_', 'UNITS', '_NOTDEF_REQUIRED_', 'SHORT_NAME', '_NOTDEF_ALLOWED_')
	FOR ii = idt[0] + 1, N_ELEMENTS(h)-1 DO BEGIN
		IF STRPOS(h[ii],'X_HEADER=') GE 0 THEN BREAK
		IF STRPOS(h[ii],'=') LE 0 THEN CONTINUE
		idr = WHERE(TAG_NAMES(time)+'(0)' EQ str_sep2(h[ii],'=', ONLY=0), cdr)
		IF cdr EQ 1 THEN BEGIN
			time.(idr[0]) = str_sep2(h[ii],'=', ONLY=1)
		ENDIF
	ENDFOR
	time.units = STRTRIM(STRCOMPRESS(time.units),2)
	wtu = STRUPCASE(STRSPLIT(time.units, ' ', COUNT=ctu, /EXTRACT))
	IF ctu LT 3 THEN RETURN, [result, 'ERROR:= time units string too short ...']
	IF ctu GT 5 THEN RETURN, [result, 'ERROR:= time units string too long ...']
	IF wtu[1] NE 'SINCE' THEN RETURN, [result, 'ERROR:=  since missing as 2nd word in time units string']
	CASE wtu[0] OF
		'MINUTES': result = [result, 'STATUS:= detected MINUTES as TIME unit']
		'HOURS': result = [result, 'STATUS:= detected HOURS as TIME unit']
		'SECONDS': result = [result, 'STATUS:= detected SECONDS as TIME unit']
		ELSE: RETURN, [result, 'ERROR:=  valid TIME unit (minutes, seconds, hours) missing']
	ENDCASE
	ref_time =  (ctu GT 3) ? string2js(wtu[2]+' '+wtu[3]) : string2js(wtu[2]+' ')
	result = [result, 'STATUS:= detected reference TIME: '+js2string(ref_time, FORM='Y-M-D H:M:S')]
ENDELSE

; create global struct
global = CREATE_STRUCT(required_global[0], '_NOTDEF_REQUIRED_')
FOR i=1, N_ELEMENTS(required_global)-1 DO $
	global = CREATE_STRUCT(global, required_global[i], '_NOTDEF_REQUIRED_')
FOR i=0, N_ELEMENTS(allowed_global)-1 DO $
	global = CREATE_STRUCT(global, allowed_global[i], '_NOTDEF_ALLOWED_')

; Loop the global section
FOR ii = idg[0] + 1, N_ELEMENTS(h)-1 DO BEGIN
	IF STRPOS(h[ii],'X_HEADER=') GE 0 THEN BREAK
	IF STRPOS(h[ii],'=') LE 0 THEN CONTINUE
	idr = WHERE(TAG_NAMES(global) EQ str_sep2(h[ii],'=', ONLY=0), cdr)
	IF cdr EQ 1 THEN BEGIN
		global.(idr[0]) = str_sep2(h[ii],'=', ONLY=1)
	ENDIF
ENDFOR

; check the global entries
tags = TAG_NAMES(global)
FOR i = 0L, N_ELEMENTS(tags) -1 DO BEGIN
	IF global.(i) EQ '_NOTDEF_REQUIRED_' THEN RETURN, [result, 'ERROR:= required info missing: '+tags[i]]
	IF global.(i) EQ '_NOTDEF_ALLOWED_' THEN result= [result, 'WARNING:= info missing: '+tags[i]]
ENDFOR
result = [result, 'STATUS:= global information processed']


; create parameter struct
param=CREATE_STRUCT('JUNK','')
idpp = WHERE(num_sd NE -2, cdpp)
FOR ii = 0L, cdpp-1 DO BEGIN
	FOR i=0L, N_ELEMENTS(required_param)-1 DO $
		param = CREATE_STRUCT(param, required_param[i]+'_'+n2s(ii)+'_', '_NOTDEF_REQUIRED_')
	FOR i=0L, N_ELEMENTS(allowed_param)-1 DO $
		param = CREATE_STRUCT(param, allowed_param[i]+'_'+n2s(ii)+'_', '_NOTDEF_ALLOWED_')
ENDFOR
param = delete_tag(param, 'JUNK')
IF _time_flag THEN param = delete_tag(param, '*_0_')


; Loop the parameter section
FOR ii = idp[0] + 1L, N_ELEMENTS(h)-1 DO BEGIN
	IF STRPOS(h[ii],'X_HEADER=') GE 0 THEN BREAK
	IF STRPOS(h[ii],'=') LE 0 THEN CONTINUE
	idr = WHERE(TAG_NAMES(param) EQ text2tagname(str_sep2(h[ii],'=', ONLY=0)), cdr)
	IF cdr EQ 1 THEN BEGIN
		param.(idr[0]) = str_sep2(h[ii],'=', ONLY=1)
	ENDIF
ENDFOR


; check the global entries
tags = TAG_NAMES(param)
FOR i = 0L, N_ELEMENTS(tags) -1 DO BEGIN
	tags[i] = replace_string(tags[i],'_', ')', POS=STRLEN(tags[i])-2)
	tags[i] = replace_string(tags[i],'_', '(', POS=STRLEN(tags[i])-5)
	IF param.(i) EQ '_NOTDEF_REQUIRED_' THEN RETURN, [result, 'ERROR:= required info missing: '+tags[i]]
	IF param.(i) EQ '_NOTDEF_ALLOWED_' THEN result= [result, 'WARNING:= info missing: '+tags[i]]
ENDFOR
result = [result, 'STATUS:= parameter information processed, please check warnings']

; data processing
d2 = d
a2 = MAKE_ARRAY(ncol,N_ELEMENTS(d), VAL=!values.d_nan, /DOUBLE)
FOR j = 0L, N_ELEMENTS(d)-1 DO BEGIN
 	IF STRTRIM(STRCOMPRESS(d[j]),2) EQ '' THEN BEGIN
    	d[j] ='!$%&?=empty' & CONTINUE & ENDIF
  	IF N_ELEMENTS(STRSPLIT(d[j],' ,;')) NE ncol THEN BEGIN
    	d[j] ='!$%&?=ncol' & CONTINUE & ENDIF
  	a = DBLARR(ncol)
    ON_IOERROR, bad_record
    READS,d[j], a
    a2[*,j] = a
    CONTINUE
bad_record:
        d[j] ='!$%&?=bad'
ENDFOR

; sort and select uniq data records
idx2 = WHERE(STRPOS(d, '!$%&?=') EQ 0, cdx2)
IF cdx2 GT 0 THEN BEGIN
	FOR ii = 0L, cdx2-1 DO $
		IF d[idx2[ii]] EQ '!$%&?=bad' THEN $
			result = [result, 'ERROR:= DATA record #'+n2s(idx2[ii])+' cannot be processed: '+ STRMID(d2[idx2[ii]], 0, 20) + '...'] $
		ELSE IF d[idx2[ii]] EQ '!$%&?=ncol' THEN $
			result = [result, 'ERROR:= DATA record #'+n2s(idx2[ii])+' num of entries does not match: '+ STRMID(d2[idx2[ii]], 0, 20) + '...'] $
		ELSE $
			result = [result, 'ERROR:= DATA record #'+n2s(idx2[ii])+' cannot be processed: ... (no data)']
	RETURN, result
ENDIF

IF _time_flag THEN BEGIN
	t_temp = a2[0,*]
	IF N_ELEMENTS(t_temp) GT 1 THEN BEGIN
  		idx = WHERE(t_temp[1:*] - t_temp  LE 0.0D, cdx)
  		IF cdx GT 0 THEN $
			RETURN, [result, 'ERROR:= '+n2s(cdx)+' DATA record(s) with non monotonic time found']
	ENDIF
ENDIF

result = [result, 'STATUS:= data successfully processed']

err = 0
RETURN, [result,'!! successfully checked at '+js2string(/NOW)+'..... (check_edf.pro)']
END
