; Copyright (c) 2006, Theo Brauers, Forschungszentrum Juelich GmbH ICG-II
; All rights reserved. Unauthorized reproduction prohibited.
; This software may be used, copied, or redistributed as long as it is not
; sold and this copyright notice is reproduced on each copy made.  This
; routine is provided as is without any express or implied warranties
; whatsoever. Email bug reports to th.brauers@fz-juelich.de
;
;+
; USERLEVEL:
;   TOPLEVEL
;
; NAME:
;   create_edf
;
; PURPOSE:
;   This procedure creates an edf file template
;
; CATEGORY:
;   DATAFILES/ZCHN
;
; CALLING SEQUENCE:
;   WRITE_ENZ, filename
;
; INPUTS:
;   Filename: output file name
;
; OPTIONAL INPUTS:
;   Struct:  data structure
;   Tagnames:   string array of tagnames to be written
;               if not present all tags will written
;
; KEYWORD PARAMETERS:
;   ERROR:  returns 0 if succesfull
;   /STDEV:  if set the stdev of all parameters (if present) will be written
;	/QUALITY: If set the quality flag of all parameters (if present) will be written
;   /INTERVAL_T:  if set the timeinterval of the time axis will be written
;   ADDITIONAL_HEADER:  if supplied this text will be added to the header of the outfile
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;   Written by: Theo Brauers 2006-Nov-14, as acopy of write_enz
;-
PRO create_edf, filename, icg_str, tagnames $
	, STDEV=stdev, INTERVAL_T=int_t $
	, N_PARAMETER = np, N_RECORDS=nr $
	, REFERENCE_DATE =reference_date, TIME_STEP=time_step $
	, LONG_NAMES = long_names $
	, UNITS=units, MISSING_VALUES=missing_values $
    , ADDITIONAL_HEADER=add_header $
	, ERROR=err, _EXTRA= e

err = 1

; some error handling
IF NOT is_string(filename) THEN RETURN
IF N_ELEMENTS(filename) NE 1 THEN RETURN

IF N_ELEMENTS(nr) NE 1 THEN nr = 5L ELSE nr = LONG(nr)

IF is_structure(icg_str) THEN BEGIN
	write_edf, filename, icgs_str, tagnames $
		, STDEV=stdev, INTERVAL_T=int_t $
    	, ADDITIONAL_HEADER=add_header $
		, ERROR=err, _EXTRA= e
	t = get_file(filename)
	idx = WHERE(STRPOS(STRTRIM(t,2), '&&&&&') EQ 0, cdx)
	IF cdx NE 1 THEN STOP
	putfile, filename, TRANSPOSE(t[0:idx+nr])
ENDIF

header1 = [ '!! TEMPLATE FOR EDF FILE !!','!! FILE '+filename, '!! CREATED ON '+SYSTIME(), $
    '!! BY '+user_at_host() ,'!! WITH '+ file_version() ,'']
IF N_ELEMENTS(add_header) GT 0 THEN header1=[header1, add_header, '']

header=[header1, 'X_HEADER=NETCDF_GLOBAL']


nl = N_ELEMENTS(long_names)
nu = N_ELEMENTS(units)

IF nl EQ 0 THEN BEGIN
	nl = N_ELEMENTS(np) NE 0 ? np < 1000 : 1
	long_names = 'LONG_NAME #'+STRTRIM(STRING(LINDGEN(nl)+1),2)
ENDIF ELSE BEGIN
	long_names = STRING(REFORM(long_names, nl))
ENDELSE

CASE nu OF
	1: units = REPLICATE(units[0], nl)
	nl: units = REFORM(units,nl)
	ELSE: units = 'UNITS #'+STRTRIM(STRING(LINDGEN(nl)+1),2)
ENDCASE

IF N_ELEMENTS(reference_date) NE 1 THEN reference_date ='2000-1-1 0:00:00'
time_step = N_ELEMENTS(time_step) NE 1 ? 3600. : time_step > 0.001
tvalue = DINDGEN(nr) * time_step
nc = 1L & text = 'TIME'
IF KEYWORD_SET(int_t) THEN BEGIN
	tvalue = [[tvalue], [MAKE_ARRAY(/DOUBLE, nr, VAL=time_step)]]
	nc += 1 & text = [text, 'INTERVAL']
ENDIF



dvalue = MAKE_ARRAY(/DOUBLE, nr, nc, /INDEX)
dvalue = [[tvalue], [dvalue]]

sd = SIZE(dvalue)
IF N_ELEMENTS(sd) EQ 5 THEN BEGIN & ix = sd[1] & iy = sd[2] & ENDIF
IF N_ELEMENTS(sd) EQ 4 THEN BEGIN & ix = 1 & iy = sd[1] & ENDIF
IF N_ELEMENTS(sd) EQ 3 THEN BEGIN & ix = 1 & iy = 1 & ENDIF

header = [header, '', 'X_HEADER=ENZ', make_enz_header(text)]
iwzn, filename, dvalue, ix, iy, header, err, _EXTRA = e

RETURN
END
