; Copyright (c) 2010, Theo Brauers, Forschungszentrum Juelich GmbH ICG-2
; All rights reserved. Unauthorized reproduction prohibited.
; This software may be used, copied, or redistributed as long as it is not
; sold and this copyright notice is reproduced on each copy made.  This
; routine is provided as is without any express or implied warranties
; whatsoever. Email bug reports to th.brauers@fz-juelich.de
;
;+
; USERLEVEL:
;   TOPLEVEL
;
; NAME:
;   write_icgs2csv
;
; PURPOSE:
;   This procedure writes an ICG structure into an csv file
;
; CATEGORY:
;   DATAFILES/ZCHN
;
; CALLING SEQUENCE:
;   write_icgs2csv, filename, struct
;
; INPUTS:
;   Struct:  data structure
;   Filename: output file name
;
; OPTIONAL INPUTS:
;   Tagnames:   string array of tagnames to be written
;               if not present all tags will written
;
; KEYWORD PARAMETERS:
;   ERROR:  returns 0 if succesfull
;   /STDEV:  if set the stdev of all parameters (if present) will be written
;   /INTERVAL_T:  if set the timeinterval of the time axis will be written
;   ADDITIONAL_HEADER:  if supplied this text will be added to the header of the outfile
;	/QUIET:	if set no info messages will be printed
;	/VERBOSE:	print all messages
; 	/REPLACE_FILLV: replac all fill values with the given fill value
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;   Written by: Theo Brauers 2010-7-20, as acopy of write_edf
;-
PRO write_icgs2csv, filename, icg_str, tagnames $
	, STDEV=stdev, INTERVAL_T=interval_t $
    , ADDITIONAL_HEADER=add_header, SEPARATOR = sep $
    , QUIET=quiet, FILL_VALUE = fill_value, REPLACE_FILLVALUE=replv $
	, ERROR=err, VERBOSE=verb, _EXTRA= e

	IF NOT KEYWORD_SET(quiet) THEN IF KEYWORD_SET(verb) THEN MESSAGE, /INFO, 'write_csv 2010-7-20'

	err = 1
	IF N_ELEMENTS(fill_value) EQ 1 THEN fill_v = DOUBLE(fill_value) ELSE fill_v = -9.9900D9

; some error handling
	IF NOT is_structure(icg_str) THEN RETURN
	IF NOT is_string(filename) THEN RETURN
	IF N_ELEMENTS(filename) NE 1 THEN RETURN
; header lines with !!
	header1 = [ '!! FILE '+filename, '!! CREATED ON '+SYSTIME(), $
    	'!! BY '+user_at_host() ,'!! WITH '+ file_version() ,'']
	IF N_ELEMENTS(add_header) GT 0 THEN header1=[header1, '!add: '+add_header, '']
; get the names
	names=get_tag_and_short_names(icg_str, count)
	IF N_ELEMENTS(tagnames) GT 0 THEN tag_n2 = tagnames ELSE tag_n2 = names.tag_name
	IF N_ELEMENTS(stdev) NE N_ELEMENTS(tag_n2) THEN $
		IF KEYWORD_SET(stdev) THEN stdev2 = MAKE_ARRAY(N_ELEMENTS(tag_n2), /BYTE, VAL=1B) $
		ELSE stdev2 = MAKE_ARRAY(N_ELEMENTS(tag_n2), /BYTE, VAL=0B) $
	ELSE stdev2 = stdev
; initialize some vars
	text='' & k=0
	iy = tag_position(icg_str, tag_n2, count)
; nothing to do
	IF count LT 1 THEN RETURN
; search time, if available
	it = tag_position(icg_str, 'TIME', ct)
	IF ct GT 1 THEN RETURN
; time column exist
	IF ct EQ 1 THEN BEGIN
    	tvalue = icg_str.(it).param
    	text = [text, 'TIME' ]
	    IF is_tag(icg_str.time, 'INTERVAL_T') AND KEYWORD_SET(interval_t) THEN BEGIN
        	tvalue = [[tvalue], [icg_str.(it).interval_t]]
        	text = [text, 'INTERVAL']
        	_form = 2
    	ENDIF ELSE _form =1
	ENDIF ELSE _form = 0
; number of rows and columns
	iy = a_not_b(a_not_b(iy, it), [-1L])
	count = N_ELEMENTS(iy)
	ixt = _form EQ 0 ? N_ELEMENTS(icg_str.(iy[0]).param) : N_ELEMENTS(tvalue)/_form
; loop data tags
	FOR i=0, count-1 DO BEGIN
    	CASE 1 OF
	        ((SIZE(icg_str.(iy[i]).param))[0] NE 1 AND N_ELEMENTS(icg_str.(iy[i]).param) NE 1): $
	            MESSAGE, 'Cannot write 3 and more dim parameters: ' + icg_str.(iy[i]).short_name, /INFO
	        (TOTAL(SIZE(icg_str.(iy[i]).param, /TYPE) EQ [0,6,7,8,9,10,11]) EQ 1): $
	            MESSAGE, 'Cannot write non num parameters: ' + icg_str.(iy[i]).short_name, /INFO
	        (N_ELEMENTS(icg_str.(iy[i]).param) NE ixt): $
	            MESSAGE, 'Cannot write params different in size: ' + icg_str.(iy[i]).short_name, /INFO
	        ELSE: BEGIN
	        	temp = DOUBLE(icg_str.(iy[i]).param)
	        	idxn = WHERE(FINITE(temp, /NAN), cdxn)
    	    	IF cdxn GT 0 THEN BEGIN
        	    	IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
            		temp[idxn] = fill_v
        		ENDIF
        		idxn = WHERE(temp EQ fill_v, cdxn)
        		IF cdxn GT 0 AND KEYWORD_SET(replv) THEN BEGIN
            		IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
            		temp[idxn] = fill_v
        		ENDIF
            	text = [text, text2tagname( icg_str.(iy[i]).short_name)]
           		dvalue = (N_ELEMENTS(dvalue) EQ 0) ? temp : [[dvalue], [temp]]
	            IF KEYWORD_SET(is_tag(icg_str.(iy[i]), 'STDEV')) AND KEYWORD_SET(stdev2[i]) THEN BEGIN
	            	temp = DOUBLE(icg_str.(iy[i]).stdev)
					idxn = WHERE(FINITE(temp, /NAN), cdxn)
	        		IF cdxn GT 0 THEN BEGIN
	            		IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
	            		temp[idxn] = fill_v
	        		ENDIF
		        	idxn = WHERE(temp EQ fill_v, cdxn)
		        	IF cdxn GT 0 AND KEYWORD_SET(replv) THEN BEGIN
		            	IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
		            	temp[idxn] = fill_v
		        	ENDIF
	           	    dvalue = [[dvalue], [temp]]
	                text = [text, 'STDEV('+text2tagname( icg_str.(iy[i]).short_name)+')']
	            ENDIF
	        END
    	ENDCASE
	ENDFOR

; crate the output array and counters
	IF N_ELEMENTS(dvalue) EQ 0 THEN dvalue = tvalue
	IF N_ELEMENTS(dvalue) GT 0 AND N_ELEMENTS(tvalue) GT 0 THEN dvalue = [[tvalue],[dvalue]]
	sd = SIZE(dvalue)
	IF N_ELEMENTS(sd) EQ 5 THEN BEGIN & ix = sd[1] & iy = sd[2] & ENDIF
	IF N_ELEMENTS(sd) EQ 4 THEN BEGIN & ix = 1 & iy = sd[1] & ENDIF
	IF N_ELEMENTS(sd) EQ 3 THEN BEGIN & ix = 1 & iy = 1 & ENDIF
; create format string from data struct
	IF is_string(sep) EQ 0 THEN sep2 = "," ELSE sep2 = sep
	CASE _form OF
		2: form1 = "((F15.3),'"+sep2+"',(F10.3),("+n2s(iy-2)+"('"+sep2+"',(G13.4))))"
		1: form1 = "((F15.3),("+n2s(iy-1)+"('"+sep2+"',(G13.4))))"
		ELSE: form1 = "((G13.4),("+n2s(iy-1)+"('"+sep2+"',(G13.4))))"
	ENDCASE
; write the data to file
	ON_IOERROR, BAD
	OPENW, unit, filename, /GET_LUN
	; PRINTF, unit, TRANSPOSE(header1)
	PRINTF, unit, arr2string(text[1:*], SEP=',')
	FOR i = 0L, ix-1 DO BEGIN
    	PRINTF, unit, dvalue[i,*], FORMAT=form1
	ENDFOR
	FREE_LUN, unit & err = 0
	RETURN
; exception handling
BAD:
	PRINT, !ERR_STRING
	FREE_LUN, unit & err = 1
	RETURN
END

